#include "defs.h"
#include "block.h"
#include "array_m.h"
#include "vector.h"
#include "hilbert.h"
#include "spins.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif
#ifdef CONFIG_MPI
#include <mpi.h>
#endif

struct vblock {     /* arguments for working threads */
  int blk;          /* block (thread) nr */
  mcplx *v0;        /* vector (used by clrvec,mltply,addav) */
  double r;         /* factor (used by mltply,addav) */
  mcplx *v1;        /* vector (used by addav) */
};

/* void-function for pthreads */
void copy_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  mcplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    v1[i]  = M_d2f(M_f2d(v0[i])*r);
    // n2t   += Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double copy_av(tvec *v1, double r, tvec *v0){  /* v1=r*v0, [return norm(v1)] v2.27 */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, copy_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#if 0
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
#endif
  return(n2);
}

/* void-function for pthreads */
void add_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  mcplx *v1=((struct vblock *)ptr)->v1+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) {
    v1[i] += M_d2f(M_f2d(v0[i])*r);
    n2t   +=  Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}

double add_av(tvec *v1, double r, tvec *v0){  /* v1+=r*v0, return norm(v1) */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    buf[blk].v1 =v1;
    parallel_start( blk, add_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
#ifdef CONFIG_MPI
  { double add=n2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    n2=sum; }
#endif
  return(n2);
}

/* void-function for pthreads */
void mltply_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0+b_ofs[blk];
  double  r=((struct vblock *)ptr)->r;
  ULong i;
  for(i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) 
    v0[i] = M_d2f(M_f2d(v0[i])*r);
  return;
}

void mltply_av(double r, tvec *v0){  /* v0*=r */
  struct vblock buf[B_NUM];
  int blk;
  for(blk=0;blk<pt_n;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0;
    parallel_start( blk, mltply_av_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  return;
}

/* void-function for pthreads */
void clrvec_block( void *ptr ){
  int  blk=((struct vblock *)ptr)->blk;
  mcplx *v=((struct vblock *)ptr)->v0+b_ofs[blk];
  ULong i;
  for (i=0l;i<b_ofs[blk+1]-b_ofs[blk];i++) v[i]=M_d2f(0);
  return;
}

void clrvec(tvec *v){ int blk;
  /* its important to inititialize memory locally in the threads for NUMA */
  struct vblock buf[B_NUM];
  if (vvv && !mpi_i) printf("\n clrvec");
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  for (blk=0;blk<pt_n;blk++) {
    if (vvv&4 && (mpi_i*pt_n+blk<4 || mpi_i*pt_n+blk>=mpi_n*pt_n-4))
       printf(" v_%d_%d[%lu]=0", mpi_i, blk, b_ofs[blk+1]-b_ofs[blk]);
    buf[blk].blk=blk;
    buf[blk].v0 =v;
    parallel_start( blk, clrvec_block, &buf[blk]);
  }
  for (blk=0;blk<pt_n;blk++) parallel_wait(blk); /* synchronize */
  if (vvv) MPI_Barrier(MPI_COMM_WORLD);
  if (vvv && !mpi_i) printf(" ok");
  return;
}

double normvec(tvec *v){ /* non-thread-parallel, ToDo: norm_v() */
  double r=0.0; ULong j;
  for (j=0l;j<b_ofs[pt_n];j++)
    r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
  if (r<=1e-20){ if(vvv&4 && !mpi_i) printf(" norm_vec<1e-20"); } else
  if (r!=1.0){ r=1.0/sqrt(r); 
    if (vvv&1 && !mpi_i) printf(" normvec 1-r*(1/r)=%.0e",1-r*(1/r));
    mltply_av(r,v); // v=r*v;
    for (r=j=0l;j<b_ofs[pt_n];j++)
      r += Norm(M_f2d(v[j]));
#ifdef CONFIG_MPI
  { double add=r, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=sum; }
#endif
    if (vvv&1 && !mpi_i) printf(" 1-norm=%.0e",1-r);
  }
  return r;
}

/* if blk==-1 save all blocks, idx>=0 index for name="d%d.vec" */
/* called only by thread blk==0 */
int savevec(tvec *v, const char *nam, int idx, int blk){
  FILE *f; ULong n2; int i=0, node; char s1[128], *s2;
  if (blk>0) {
    printf("\nWarning savevec: code called from working thread, skip");
    return 0;
  }
  if (idx>=0) i=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
  if (i<0) E0("file name to long");
  if ((vvv&1) && !mpi_i) printf(" save_v %s",s2);
  for (node=0; node<mpi_n; node++) {
    MPI_Barrier(MPI_COMM_WORLD);  // synchronize threads
    if (mpi_i!=node) continue;
    if (!(f=fopen(s2,((node)?"ab":"wb")))) { E1("fopen(%s)",s2); return 1;}
    if (b_ofs[pt_n]) {
      if(!v){ E0("savevec failed, v==0"); return 3;}
      n2=fwrite((void *)v,Csize,(size_t)b_ofs[pt_n],f); /* full node */
      if( n2!=b_ofs[pt_n] ){ E0("fwrite");return 2; }
    } fclose(f);
  }
  return 0;
}

int loadvec(tvec *v,const char *nam,int idx,int blk){
  FILE *f; ULong n2; int i=0; char s1[128], *s2;
  if (mpi_n>1) {
    if (!mpi_i) printf("\nWarning loadvec: missing MPI code, skip");
    return 0;
  }
  if (idx>=0) i=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
  if (i<0) E0("file name to long");
  if (vvv&1) printf(" load_v %s",s2);
  if (!(f=fopen(s2,"rb"))) { E0("fopen");return 1;}
  if (b_ofs[pt_n]) {
    if(!v){ E0("loadvec failed, v==0"); return 3;}
    n2=fread((void *)v,Csize,(size_t)b_ofs[pt_n],f);
    if( n2!=b_ofs[pt_n] ){ E0("fread");return 2; }
  } fclose(f);
  return 0;
}

int loadvec2(tvec *v,const char *nam,int idx){ /* load v from textfile */
  FILE *f; double r; ULong j; tbase i; int k,l,ret; char buf[255];
  if (mpi_n>1) F0("missing MPI code");
  int ii=0; char s1[128],*s2;
  if(idx>=0) ii=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
  if(ii<0) E0("file name to long");
  if(!v){ printf(" failed, v==NULL"); return 3;}
  if(!(f=fopen(s2,"rt"))){E1("loadvec_fopen(%s)",s2);return 1;}
  clrvec(v);ret=1;
  while(!feof(f)){
      if(!fgets(buf,255,f))break; /* zeile lesen: 0.25 |ud.3u> /n */
      if(sscanf(buf,"%lf",&r)!=1)continue;
      for(k=0;buf[k];k++) if(buf[k]=='|')break;
      i=0l;j=0l;l=2*(nn-1) CM(/2,+0,+0);
      if(buf[k]=='|'){
         for(k++;buf[k]!=0 && l>=0;k++,CM(l--,l-=2,l-=2))	/* Apr99 */
         if(buf[k]=='.'){                   } else
         if(buf[k]=='u'){i|=CM(0l,1l,1l)<<l;} else
         if(buf[k]=='d'){i|=CM(1l,2l,2l)<<l;} else
         if(buf[k]=='3'){i|=         3l <<l;} else
         E0("unknown symbol (not .ud3)"); }
      else E1("no config |xxxx> found in %s",nam);   
      j=scfg2idx(i); /* 0 .. node_n1-1 */
      if(vvv)printf("\n read: %6.3f |%s> => %ld",r,ccfg(i),j);
      v[j]=M_d2f(r);ret=0;}
  fclose(f);
  return ret;
}
   
#ifndef RAND_MAX
#define RAND_MAX 0x7fff
#endif
extern int startvec;
/* --- last change: 7.12.98 JS bug! --- 
 *  sum v[] during setting for faster normalization???
 *  ini could be choosen randomly to ensure different numeric errors
 */
void randvec(tvec *v, int ini){
    ULong i;

    ini += (startvec&~7);
    if (vvv && !mpi_i) printf(" randvec%d-%d", startvec&7, ini);
    if (vvv) MPI_Barrier(MPI_COMM_WORLD);
    clrvec(v); /* memory first touch for ccNUMA */
    if (startvec==1){ /* set one component to 1, rest to 0 */
      if (  (ULong)ini>=mpi_ofs[mpi_i  ]
         && (ULong)ini< mpi_ofs[mpi_i+1]  )
        v[  (ULong)ini -mpi_ofs[mpi_i  ] ]=M_d2f(1); } else
    if (startvec==2){ loadvec( v,FileGS,ini,-1);   } else
    if (startvec==3){ loadvec2(v,FileSV,ini);      } else
    if (startvec==5){ /* ToDo: set all v[i]=1 ? */
      E0("startvec=5 removed in v2.35, ToDo: storeh may return i_min");
    } else
    if (startvec>4){ printf(" no startvec"); } else
    { /* srand(rnd);rnd++; -- bad idea, resulting in a bug until v2.17 */ 
      /* ToDo: spread in a way that norm is exactly 1.0, use l^2+m^2=n^2 */
      srand(ini);
      for(i=0l;i<b_ofs[pt_n];i++)
        v[i]=M_d2f((1023 & rand())/1024.); /* Jul00 */
      if (!mpi_i) v[0]=M_d2f(0.5); /* be sure v is never zero */
    }
    if (vvv) MPI_Barrier(MPI_COMM_WORLD);
    if (vvv && !mpi_i) printf(" randvec=ok normvec");
    normvec(v); return;
}
    
/* scalar product, need correct order: <a+ib|c+id>!=<c+id|a+ib> */
cplx sprod(tvec *a, tvec *b){ /* non-thread-parallel */
    cplx r=0.0; ULong i;
    for (i=0l;i<b_ofs[pt_n];i++) 
        r+=Conj(M_f2d(a[i])) * M_f2d(b[i]);
#ifdef CONFIG_MPI
  { double add[2]={Re(r),Im(r)}, sum[2]={0,0};
    MPI_Allreduce(&add, &sum, (VecType&4)/4+1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r=C(sum[0],sum[1]); }
#endif
    return(r);
}
    
int  iortho(tvec *v0, tvec *v1){ /* 0=ok,1=parallel */
    cplx r1; double r2; ULong i;
    r1=sprod(v0,v1);
    if (Norm(r1)<1.0e-10) return(0);
    r2=0.0;
    for (i=0l;i<b_ofs[pt_n];i++){
      /* Conj(r1) for correct degeneracy for complex vectors */
      v0[i]-=M_d2f(Conj(r1)*M_f2d(v1[i])); /* round error ??? */
      r2+=Norm(M_f2d(v0[i]));
    }
#ifdef CONFIG_MPI
  { double add=r2, sum=0;
    MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    r2=sum; }
#endif
    /* 0.995 was successfull for degeneracy of kagome-JJ N=54 u+d=51+3 q_0
     * 0.99999 results in to high degeneracy (909 instead of 72) Jul04
     * */
    if (Norm(r1)>0.995) {printf(" iortho: v0=v1");return(1);}
    r2=1.0/sqrt(r2);
//    for (i=0l;i<b_ofs[pt_n];i++)
//      v0[i]=M_d2f(M_f2d(v0[i])*r2);  /* call mltly_av ??? */
    mltply_av(r2,v0);
    return(0);
}

